import React, { FunctionComponent } from 'react';
import styles from 'EnhancedTable.module.scss';
import { TableInstance } from 'react-table';
import arrayMove from 'array-move';
import {
  SortableContainer,
  SortableContainerProps,
  SortableElement,
  SortableElementProps,
  SortableHandle,
} from 'react-sortable-hoc';
import clsx from 'clsx';

export const DraggableContainer = SortableContainer(
  ({ children }: { children: React.ReactElement } & SortableContainerProps) =>
    children
);
export const DraggableElement = SortableElement(
  ({ children }: { children: React.ReactElement } & SortableElementProps) =>
    children
);

const DragHandle = SortableHandle(
  ({ children }: { children: React.ReactElement }) => children
);

type Props = {
  instance: TableInstance<any>;
  stableColumns: string[];
};

const TableHeader: FunctionComponent<Props> = (props) => {
  const { instance, stableColumns } = props;
  const { allColumns, visibleColumns, setColumnOrder, headerGroups } = instance;

  function findGlobalIndex(index: number) {
    return allColumns.findIndex((item) => {
      return visibleColumns[index].id === item.id;
    });
  }

  function getNewStickyIndexForStable(newSticky: 'left' | 'right') {
    const newLeftStickyIndex = visibleColumns.filter((item) => item.isStable)
      .length;

    const newRightStickyIndex = visibleColumns.findIndex(
      (item) => item.sticky === 'right'
    );

    return newSticky === 'left' ? newLeftStickyIndex : newRightStickyIndex - 1;
  }

  function getNewStickyIndex(oldIndex: number, newIndex: number) {
    const oldSticky = visibleColumns[oldIndex].sticky;
    const newSticky = visibleColumns[newIndex].sticky;

    //move from normal column to sticky column
    if (!oldSticky && newSticky) {
      if (stableColumns.includes(visibleColumns[newIndex].id)) {
        newIndex = getNewStickyIndexForStable(newSticky);
      }

      visibleColumns[oldIndex].setSticky(newSticky);
    }

    //move from sticky column to normal column
    if (oldSticky && !newSticky) {
      visibleColumns[oldIndex].setSticky();
    }

    return newIndex;
  }

  const onSortEnd = ({
    oldIndex,
    newIndex,
  }: {
    oldIndex: number;
    newIndex: number;
  }) => {
    newIndex = getNewStickyIndex(oldIndex, newIndex);

    if (
      stableColumns.includes(visibleColumns[oldIndex].id) ||
      stableColumns.includes(visibleColumns[newIndex].id)
    )
      return;

    const globalOldIndex = findGlobalIndex(oldIndex);
    if (globalOldIndex === -1) return;
    const globalNewIndex = findGlobalIndex(newIndex);
    if (globalNewIndex === -1) return;

    setColumnOrder(
      arrayMove(allColumns, globalOldIndex, globalNewIndex).map(
        (item) => item.id
      )
    );
  };
  return (
    <div
      className={styles.header}
      style={{ width: instance.totalVisibleColumnsWidth }}
    >
      {headerGroups.map((headerGroup, index) => (
        <DraggableContainer
          onSortEnd={onSortEnd}
          key={index}
          axis="x"
          distance={5}
          useDragHandle
        >
          <div className={styles.tr} {...headerGroup.getHeaderGroupProps()}>
            {headerGroup.headers
              .filter((column) => !column.isHidden)
              .map((column, index) => {
                return (
                  <DraggableElement
                    key={column.id}
                    index={index}
                    disabled={stableColumns.includes(column.id)}
                  >
                    <div
                      className={clsx(styles.th, 'th')}
                      {...column.getHeaderProps()}
                    >
                      <DragHandle>
                        <span
                          className={styles.headerCell}
                          {...column.getSortByToggleProps()}
                        >
                          <span>{column.render('Header')}</span>
                          {column.canSort && (
                            <span className={styles.sortColumn}>
                              {column.isSorted ? (
                                column.isSortedDesc ? (
                                  <span className={styles.down} />
                                ) : (
                                  <span className={styles.up} />
                                )
                              ) : (
                                <React.Fragment>
                                  <span className={styles.up} />
                                  <span className={styles.down} />
                                </React.Fragment>
                              )}
                            </span>
                          )}
                        </span>
                      </DragHandle>

                      {column.canResize && (
                        <div
                          {...column.getResizerProps()}
                          className={clsx(
                            styles.resizer,
                            column.isResizing && styles.resizing
                          )}
                        />
                      )}
                    </div>
                  </DraggableElement>
                );
              })}
          </div>
        </DraggableContainer>
      ))}
    </div>
  );
};

export default TableHeader;
