import {
  actions,
  Column as IColumn,
  Hooks,
  TableInstance,
  TableState,
} from 'react-table';
import React, { useEffect } from 'react';
import { SELECTION_COLUMN_ID } from 'constants/constants';

/**
 * Modifying table plugin from https://github.com/GuillaumeJasmin/react-table-sticky
 * Adding function setSticky to column in order to let column can change sticky property in run time
 */
interface Column {
  Header: any;
  columns?: Column[];
  getHeaderProps: () => {
    style: object;
  };
  id: string | number;
  parent?: Column;
  sticky?: 'left' | 'right';
  totalLeft: number;
}

// sticky action
const SET_STICKY_COLUMN_ACTION = 'SET_STICKY_TABLE_COLUMN';
const RESET_STICKY_COLUMNS_ACTION = 'RESET_STICKY_COLUMNS_ACTION';

export const checkErrors = (columns: Column[]) => {
  const hasGroups = !!columns.find((column: Column) => column.parent);
  const stickyColumnsWithoutGroup = columns
    .filter((column: Column) => column.sticky && !column.parent)
    .map(({ Header }) => `'${Header}'`);

  if (hasGroups && stickyColumnsWithoutGroup.length) {
    throw new Error(`WARNING react-table-sticky:
      \nYour ReactTable has group and sticky columns outside groups, and that will break UI.
      \nYou must place ${stickyColumnsWithoutGroup.join(
        ' and '
      )} columns into a group (even a group with an empty Header label)\n`);
  }

  const bugWithUnderColumnsSticky = columns.find(
    (parentCol) =>
      !parentCol.sticky &&
      parentCol.columns &&
      parentCol.columns.find((col) => col.sticky)
  );

  if (!bugWithUnderColumnsSticky) return;

  // @ts-ignore
  const childBugs = bugWithUnderColumnsSticky.columns.find(
    ({ sticky }) => sticky
  );

  if (!childBugs) return;

  throw new Error(`WARNING react-table-sticky:
    \nYour ReactTable contain columns group with at least one child columns sticky.
    \nWhen ReactTable has columns groups, only columns groups can be sticky
    \nYou must set sticky: 'left' | 'right' for the '${bugWithUnderColumnsSticky.Header}'
    column, or remove the sticky property of '${childBugs.Header}' column.`);
};

export function getStickyValue(column: Column): null | 'left' | 'right' {
  if (column.sticky === 'left' || column.sticky === 'right') {
    return column.sticky;
  }

  if (column.parent) {
    return getStickyValue(column.parent);
  }

  return null;
}

export function columnIsLastLeftSticky(
  columnId: Column['id'],
  columns: any
): boolean {
  const index = columns.findIndex(({ id }: any) => id === columnId);
  const column = columns[index];
  const nextColumn = columns[index + 1];
  const columnIsLeftSticky = getStickyValue(column) === 'left';
  const nextColumnIsLeftSticky =
    nextColumn && getStickyValue(nextColumn) === 'left';
  return columnIsLeftSticky && !nextColumnIsLeftSticky;
}

export function columnIsFirstRightSticky(
  columnId: Column['id'],
  columns: any
): boolean {
  const index = columns.findIndex(({ id }: any) => id === columnId);
  const column = columns[index];
  const prevColumn = columns[index - 1];
  const columnIsRightSticky = getStickyValue(column) === 'right';
  const prevColumnIsRightSticky =
    prevColumn && getStickyValue(prevColumn) === 'right';
  return columnIsRightSticky && !prevColumnIsRightSticky;
}

export function getMarginRight(columnId: Column['id'], columns: any) {
  const currentIndex = columns.findIndex(({ id }: any) => id === columnId);
  let rightMargin = 0;
  for (let i = currentIndex + 1; i < columns.length; i += 1) {
    if (columns[i].isVisible !== false) {
      rightMargin += columns[i].width;
    }
  }

  return rightMargin;
}

const cellStylesSticky = {
  // hard coded inline style will be remove in the next major release
  position: 'sticky',
  zIndex: 3,
};

function findHeadersSameLevel(header: any, headers: any) {
  return headers.filter((flatHeaderItem: any) => {
    return flatHeaderItem.depth === header.depth;
  });
}

function getStickyProps(header: Column, instance: TableInstance<any>) {
  let style = {};
  const dataAttrs = {};

  checkErrors(instance.columns as Column[]);

  const sticky = getStickyValue(header);

  if (sticky) {
    style = {
      ...cellStylesSticky,
    };

    // @ts-ignore
    dataAttrs['data-sticky-td'] = true;

    const headers = findHeadersSameLevel(header, instance.flatHeaders);

    const margin =
      sticky === 'left' ? header.totalLeft : getMarginRight(header.id, headers);

    style = {
      ...style,
      [sticky]: `${margin}px`,
    };

    const isLastLeftSticky = columnIsLastLeftSticky(header.id, headers);

    if (isLastLeftSticky) {
      // @ts-ignore
      dataAttrs['data-sticky-last-left-td'] = true;
    }

    const isFirstRightSticky = columnIsFirstRightSticky(header.id, headers);

    if (isFirstRightSticky) {
      // @ts-ignore
      dataAttrs['data-sticky-first-right-td'] = true;
    }
  }

  return {
    style,
    ...dataAttrs,
  };
}

/**
 * Modified part
 * Add function setSticky for column
 */
function useInstance(instance: TableInstance) {
  const {
    tableKey,
    flatHeaders,
    dispatch,
    state: { leftStickyColumns, rightStickyColumns },
  } = instance;

  const resetAllStickyColumns = React.useCallback(
    () => dispatch({ type: RESET_STICKY_COLUMNS_ACTION }),
    [dispatch]
  );

  const setStickyColumn = React.useCallback(
    (value, columnId) =>
      dispatch({
        type: SET_STICKY_COLUMN_ACTION,
        columnId: columnId,
        value,
      }),
    [dispatch]
  );

  useEffect(() => {
    localStorage.setItem(
      tableKey + '_leftStickyColumns',
      JSON.stringify(leftStickyColumns)
    );
  }, [leftStickyColumns, tableKey]);

  useEffect(() => {
    localStorage.setItem(
      tableKey + '_rightStickyColumns',
      JSON.stringify(rightStickyColumns)
    );
  }, [rightStickyColumns, tableKey]);

  flatHeaders.forEach((column, index) => {
    const isLeft = leftStickyColumns.includes(column.id);
    const isRight = rightStickyColumns.includes(column.id);
    column.sticky = isLeft ? 'left' : isRight ? 'right' : undefined;
    column.setSticky = (value) => {
      if (value !== column.sticky) setStickyColumn(value, column.id);
    };
  });

  Object.assign(instance, {
    resetAllStickyColumns,
  });
}

/**
 * Modified part
 * Handle sticky state in table
 */
function reducer(
  state: TableState,
  action: any,
  previousState: TableState,
  instance: TableInstance
) {
  function getDefaultStickyColumns(type: 'left' | 'right') {
    const columns = instance.columns as IColumn[];
    const defaultColumns = columns
      .filter((column) => column.sticky === type)
      .map((column) => {
        return column.id || column.accessor || column.Header;
      });

    if (type === 'left') defaultColumns.push(SELECTION_COLUMN_ID);
    return defaultColumns;
  }

  if (action.type === actions.init) {
    const getStickyColumns = (type: 'left' | 'right') => {
      const storageStickyColumns: string[] = localStorage.getItem(
        `${instance.tableKey}_${type}StickyColumns`
      )
        ? JSON.parse(
            localStorage.getItem(
              `${instance.tableKey}_${type}StickyColumns`
            ) as string
          )
        : undefined;
      if (
        Array.isArray(storageStickyColumns) &&
        storageStickyColumns.length > 0
      )
        return storageStickyColumns;

      return getDefaultStickyColumns(type);
    };
    const leftStickyColumns = getStickyColumns('left');

    return {
      ...state,
      leftStickyColumns: leftStickyColumns.includes(SELECTION_COLUMN_ID)
        ? leftStickyColumns
        : [SELECTION_COLUMN_ID, ...leftStickyColumns],
      rightStickyColumns: [...getStickyColumns('right')],
    };
  }

  if (action.type === SET_STICKY_COLUMN_ACTION) {
    switch (action.value) {
      case 'left':
        return {
          ...state,
          leftStickyColumns: [...state.leftStickyColumns, action.columnId],
          rightStickyColumns: state.rightStickyColumns.filter(
            (item) => item !== action.columnId
          ),
        };
      case 'right':
        return {
          ...state,
          rightStickyColumns: [...state.rightStickyColumns, action.columnId],
          leftStickyColumns: state.leftStickyColumns.filter(
            (item) => item !== action.columnId
          ),
        };
      // remove sticky for column
      default:
        return {
          ...state,
          leftStickyColumns: state.leftStickyColumns.filter(
            (item) => item !== action.columnId
          ),
          rightStickyColumns: state.rightStickyColumns.filter(
            (item) => item !== action.columnId
          ),
        };
    }
  }

  if (action.type === RESET_STICKY_COLUMNS_ACTION) {
    return {
      ...state,
      leftStickyColumns: getDefaultStickyColumns('left'),
      rightStickyColumns: getDefaultStickyColumns('right'),
    };
  }
}

export const useSticky = (hooks: Hooks<any>) => {
  // modified part
  hooks.useInstance.push(useInstance);
  // modified part
  // @ts-ignore
  hooks.stateReducers.push(reducer);

  hooks.getHeaderProps.push((props: any, { instance, column }: any) => {
    const nextProps = getStickyProps(column, instance);
    return [props, nextProps];
  });

  hooks.getCellProps.push((props: any, { instance, cell }: any) => {
    const nextProps = getStickyProps(cell.column, instance);
    return [props, nextProps];
  });
};

useSticky.pluginName = 'useSticky';
