import React, { ChangeEvent, FunctionComponent } from 'react';
import { ColumnInstance, TableInstance } from 'react-table';
import { Checkbox, FormControlLabel, Popover } from '@material-ui/core';
import styles from 'EnhancedTable.module.scss';
import clsx from 'clsx';
import { AppColumnsIcon, AppDownIcon } from 'assets/icons/AppIcon';
import { SELECTION_COLUMN_ID } from 'constants/constants';
import isEqual from 'lodash/isEqual';

type Props = {
  instance: TableInstance<any>;
  stableColumns: Array<string>;
  defaultHiddenColumns?: string[];
};
const ColumnDisplay: FunctionComponent<Props> = (props) => {
  const { stableColumns, instance, defaultHiddenColumns } = props;
  const { allColumns, state, setColumnOrder } = props.instance;
  const { columnOrder, leftStickyColumns } = state;
  const [anchorEl, setAnchorEl] = React.useState<HTMLButtonElement | null>(
    null
  );

  const handleClick = (event: React.MouseEvent<HTMLButtonElement>) => {
    setAnchorEl(event.currentTarget);
  };

  const handleClose = () => {
    setAnchorEl(null);
  };

  function isDisableColumn(column: ColumnInstance<any>) {
    return stableColumns.includes(column.id) || !column.Header;
  }

  const isShowAll = instance.visibleColumns.length === allColumns.length;
  function toggleAll(event: ChangeEvent<HTMLInputElement>) {
    if (event.target.checked) {
      instance.setHiddenColumns([]);
    } else {
      instance.setHiddenColumns(
        allColumns
          .map((item) => item.id)
          .filter((index) => !stableColumns.includes(index))
      );
    }
  }

  function priorityLeftStickyReorder() {
    let newColumnOrder = Array.from(
      new Set([SELECTION_COLUMN_ID, ...columnOrder, ...leftStickyColumns])
    );

    newColumnOrder = newColumnOrder.sort(
      (a, b) =>
        Number(leftStickyColumns.includes(b)) -
        Number(leftStickyColumns.includes(a))
    );

    if (!isEqual(newColumnOrder, columnOrder)) {
      setColumnOrder(newColumnOrder);
    }
  }

  const open = Boolean(anchorEl);
  return (
    <React.Fragment>
      <span className={styles.action} onClick={handleClick}>
        <AppColumnsIcon className={styles.iconColumns} />
        <span>COLUMNS</span>
        <AppDownIcon className={styles.iconDown} />
      </span>
      <Popover
        open={open}
        anchorEl={anchorEl}
        onClose={handleClose}
        anchorOrigin={{
          vertical: 'bottom',
          horizontal: 'right',
        }}
        transformOrigin={{
          vertical: 'top',
          horizontal: 'left',
        }}
      >
        <div className={styles.popover}>
          <div>
            <FormControlLabel
              title={'Toggle column visibility'}
              control={
                <Checkbox
                  color="primary"
                  checked={isShowAll}
                  indeterminate={!isShowAll && !instance.allColumnsHidden}
                  onChange={(event) => {
                    toggleAll(event);
                    if (event.target.checked) {
                      priorityLeftStickyReorder();
                    }
                  }}
                />
              }
              label={<span className="primary-text">All</span>}
            />
            <span
              className={styles.clearOrderTag}
              onClick={() => {
                instance.setColumnOrder([]);
                instance.resetAllStickyColumns();
                if (
                  Array.isArray(defaultHiddenColumns) &&
                  defaultHiddenColumns.length > 0
                ) {
                  instance.setHiddenColumns(defaultHiddenColumns);
                }
              }}
              title={'Reset the order of table columns'}
            >
              Reset Order
            </span>
          </div>
          {allColumns
            .filter((column) => !column.isHidden)
            .map((column) => (
              <div key={column.id}>
                <FormControlLabel
                  className={clsx(isDisableColumn(column) && 'displayNone')}
                  control={
                    <Checkbox
                      {...column.getToggleHiddenProps()}
                      checked={column.isVisible}
                      onChange={(event) => {
                        column.toggleHidden();
                        if (event.target.checked) {
                          priorityLeftStickyReorder();
                        }
                      }}
                      color="primary"
                    />
                  }
                  label={
                    typeof column.Header === 'string'
                      ? column.Header
                      : column.id
                  }
                />
              </div>
            ))}
          <br />
        </div>
      </Popover>
    </React.Fragment>
  );
};

export default React.memo(ColumnDisplay);
